from web3 import Web3

from defi_services.abis.lending.justlend.just_token_abi import JUST_TOKEN_ABI
from defi_services.abis.lending.justlend.justlend_comptroller_abi import JUSTLEND_COMPTROLLER_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.lending_constant import Lending
from defi_services.constants.token_constant import Token
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.lending.compound_service import CompoundStateService
from defi_services.services.lending.lending_info.tron.justlend_tron import JUSTLEND_TRON


class JustLendInfo:
    mapping = {
        Chain.tron: JUSTLEND_TRON
    }


class JustLendStateService(CompoundStateService):
    def __init__(self, state_service: StateQuerier, chain_id: str = "0x2b6653dc"):
        super().__init__(state_service, chain_id)
        self.name = f"{chain_id}_{Lending.justlend}"
        self.chain_id = chain_id
        self.pool_info = JustLendInfo.mapping.get(chain_id)
        self.comptroller_abi = JUSTLEND_COMPTROLLER_ABI

        # BASIC FUNCTIONS

    def get_service_info(self):
        info = {
            Lending.justlend: {
                "chain_id": self.chain_id,
                "type": "lending",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_dapp_asset_info(
            self,
            block_number: int = "latest"):
        _w3 = self.state_service.get_w3()
        comptroller_contract = _w3.eth.contract(
            address=_w3.to_checksum_address(self.pool_info.get("comptrollerAddress")), abi=self.comptroller_abi)
        ctokens = []
        for token in comptroller_contract.functions.getAllMarkets().call(block_identifier=block_number):
            ctokens.append(token)

        tokens = [Web3.to_checksum_address(i) for i in ctokens]
        reserves_info = {}
        queries = {}
        for token in tokens:
            key = f"underlying_{token}_latest".lower()
            queries[key] = {
                "address": token,
                "abi": JUST_TOKEN_ABI,
                "params": [],
                "function": "underlying",
                "block_number": "latest"
            }

            exchange_rate_query_id = f'exchangeRateStored_{token}_{block_number}'
            queries[exchange_rate_query_id] = self.get_ctoken_function_info(
                ctoken=token, fn_name='exchangeRateStored', block_number=block_number)

            markets = f"markets_{token}_latest".lower()
            queries[markets] = self.get_comptroller_function_info("markets", [token])
        decoded_data = self.state_service.query_state_data(queries)
        for token in tokens:
            key = f"underlying_{token}_latest".lower()
            markets = f"markets_{token}_latest".lower()
            underlying = decoded_data.get(key).lower()
            liquidation_threshold = decoded_data.get(markets)[1] / 10 ** 18

            if underlying != Token.native_token:
                underlying_contract = _w3.eth.contract(address=Web3.to_checksum_address(underlying), abi=ERC20_ABI)
                underlying_decimal = underlying_contract.functions.decimals().call()
            else:
                underlying_decimal = Chain.native_decimals.get(self.chain_id, 18)
            exchange_rate_query_id = f'exchangeRateStored_{token}_{block_number}'
            exchange_rate = decoded_data.get(exchange_rate_query_id) / 10 ** (18 - 8 + underlying_decimal)

            reserves_info[underlying] = {
                'cToken': token.lower(),
                "exchangeRate": exchange_rate,
                "liquidationThreshold": liquidation_threshold,
                "loanToValue": liquidation_threshold
            }
        return reserves_info

    # REWARDS BALANCE
    def get_rewards_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict = None,
            block_number: int = "latest",
    ):
        rpc_call = self.get_comptroller_function_info("compAccrued", [wallet])
        get_reward_id = f"compAccrued_{self.name}_{wallet}_{block_number}".lower()
        return {get_reward_id: rpc_call}

    def calculate_rewards_balance(
            self, wallet: str, reserves_info: dict, decoded_data: dict, block_number: int = "latest"):
        get_reward_id = f"compAccrued_{self.name}_{wallet}_{block_number}".lower()
        rewards = decoded_data.get(get_reward_id) / 10 ** 18
        reward_token = self.pool_info.get("rewardToken")
        result = {
            reward_token: {"amount": rewards}
        }
        return result

    # WALLET DEPOSIT BORROW BALANCE
    def get_wallet_deposit_borrow_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict,
            block_number: int = "latest",
            health_factor: bool = False
    ):

        rpc_calls = {}

        # Check asset is collateral
        assets_in_query_id = f"getAssetsIn_{self.pool_info['comptrollerAddress']}_{wallet}_{block_number}".lower()
        rpc_calls[assets_in_query_id] = self.get_comptroller_function_info(
            fn_name='getAssetsIn', fn_paras=[wallet], block_number=block_number)

        for token, value in reserves_info.items():
            underlying = token
            ctoken = value.get('cToken')
            if token == Token.native_token:
                underlying = Token.wrapped_token.get(self.chain_id)
            underlying_borrow_key = f"borrowBalanceCurrent_{ctoken}_{wallet}_{block_number}".lower()
            underlying_balance_key = f"balanceOfUnderlying_{ctoken}_{wallet}_{block_number}".lower()
            underlying_decimals_key = f"decimals_{underlying}_{block_number}".lower()
            rpc_calls[underlying_borrow_key] = self.get_ctoken_function_info(
                ctoken, "borrowBalanceCurrent", [wallet])
            rpc_calls[underlying_balance_key] = self.get_ctoken_function_info(
                ctoken, "balanceOfUnderlying", [wallet])
            rpc_calls[underlying_decimals_key] = self.state_service.get_function_info(
                underlying, ERC20_ABI, "decimals", []
            )

        return rpc_calls

    def calculate_wallet_deposit_borrow_balance(
            self,
            wallet: str,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict = None,
            pool_decimals: int = 18,
            block_number: int = "latest",
            health_factor: bool = False
    ):
        assets_in_query_id = f"getAssetsIn_{self.pool_info['comptrollerAddress']}_{wallet}_{block_number}".lower()
        assets_in = [t.lower() for t in decoded_data[assets_in_query_id]]

        if token_prices is None:
            token_prices = {}
        result = {}
        total_borrow = 0
        total_collateral = 0
        for token, value in reserves_info.items():
            data = {}
            underlying = token
            ctoken = value.get("cToken")
            if token == Token.native_token:
                underlying = Token.wrapped_token.get(self.chain_id)
            get_total_deposit_id = f"balanceOfUnderlying_{ctoken}_{wallet}_{block_number}".lower()
            get_total_borrow_id = f"borrowBalanceCurrent_{ctoken}_{wallet}_{block_number}".lower()
            get_decimals_id = f"decimals_{underlying}_{block_number}".lower()
            decimals = decoded_data[get_decimals_id]
            deposit_amount = decoded_data[get_total_deposit_id] / 10 ** decimals
            borrow_amount = decoded_data[get_total_borrow_id] / 10 ** decimals
            data[token] = {
                "borrow_amount": borrow_amount,
                "deposit_amount": deposit_amount,
                "is_collateral": ctoken in assets_in
            }

            if token_prices:
                token_price = token_prices.get(underlying)
            else:
                token_price = None
            if token_price is not None:
                deposit_amount_in_usd = deposit_amount * token_price
                borrow_amount_in_usd = borrow_amount * token_price
                data[token]['borrow_amount_in_usd'] = borrow_amount_in_usd
                data[token]['deposit_amount_in_usd'] = deposit_amount_in_usd
                total_borrow += borrow_amount_in_usd
                if data[token]['isCollateral']:
                    total_collateral += deposit_amount_in_usd * value.get("liquidationThreshold")

            result[ctoken] = data
        if health_factor:
            if total_collateral and total_borrow:
                result['health_factor'] = total_collateral/total_borrow
            elif total_collateral:
                result['health_factor'] = 100
            else:
                result['health_factor'] = 0
        return result



import logging

from defi_services.constants.query_constant import Query
from defi_services.constants.token_constant import Token

logger = logging.getLogger("Dex Protocol State Service")


class DexProtocolServices:
    # BASIC FUNCTIONS
    def get_service_info(self):
        # Get basic information of service
        return {}

    def get_function_info(
            self,
            query_types: list,
            wallet: str = None,
            block_number: int = "latest",
            **kwargs
    ):
        rpc_calls = {}
        supplied_data = kwargs.get("supplied_data", {})

        if Query.lp_token_list in query_types:
            limit_ = kwargs.get("number_lp")
            rpc_calls.update(self.get_all_supported_lp_token(limit_, supplied_data))

        if Query.farming_lp_token_list in query_types:
            limit_ = kwargs.get("number_lp")
            rpc_calls.update(self.get_farming_supported_lp_token(limit_))
        if Query.important_lp_token_list in query_types:
            rpc_calls.update(self.get_important_lp_token(supplied_data, block_number))

        if Query.lp_token_info in query_types:
            rpc_calls.update(self.get_lp_token_function_info(supplied_data, block_number))

        if Query.token_pair_balance in query_types:
            rpc_calls.update(self.get_balance_of_token_function_info(supplied_data, block_number))

        if Query.dex_user_nft in query_types and wallet and wallet != Token.native_token:
            rpc_calls.update(self.get_all_nft_token_of_user_function(wallet, block_number))

        if Query.dex_user_info in query_types and wallet and wallet != Token.native_token:
            stake = kwargs.get("stake", False)
            rpc_calls.update(self.get_user_info_function(wallet, supplied_data, stake, block_number))

        if Query.dex_user_token_balance in query_types:
            rpc_calls.update(self.get_user_token_amount_function(wallet, supplied_data, block_number))

        if Query.protocol_reward in query_types and wallet and wallet != Token.native_token:
            rpc_calls.update(self.get_rewards_balance_function_info(wallet, supplied_data, block_number))

        return rpc_calls

    def get_data(
            self,
            query_types: list,
            wallet: str,
            decoded_data: dict,
            block_number: int = 'latest',
            **kwargs
    ):
        result = {}
        supplied_data = kwargs.get("supplied_data", {})
        if Query.lp_token_list in query_types:
            limit_ = kwargs.get("number_lp")
            result.update(self.decode_all_supported_lp_token(limit_, decoded_data, supplied_data))

        if Query.farming_lp_token_list in query_types:
            result.update(self.decode_farming_supported_lp_token(decoded_data))

        if Query.important_lp_token_list in query_types:
            result.update(self.decode_important_lp_token(supplied_data, decoded_data))

        if Query.lp_token_info in query_types:
            result.update(self.decode_lp_token_info(supplied_data, decoded_data, block_number))

        if Query.token_pair_balance in query_types:
            result.update(self.decode_balance_of_token_function_info(
                supplied_data, decoded_data, block_number))

        if Query.dex_user_nft in query_types and wallet and wallet != Token.native_token:
            result.update(self.decode_all_nft_token_of_user_function(decoded_data))

        if Query.dex_user_info in query_types and wallet and wallet != Token.native_token:
            stake = kwargs.get("stake", False)
            result.update(self.decode_user_info_function(wallet, supplied_data, decoded_data, stake, block_number))

        if Query.protocol_reward in query_types and wallet and wallet != Token.native_token:
            result.update(self.calculate_rewards_balance(wallet, supplied_data, decoded_data, block_number))

        if Query.dex_user_token_balance in query_types:
            result.update(self.decode_user_token_amount_function(wallet, supplied_data, decoded_data, block_number))

        return result

    # Lp token function
    def get_lp_token_function_info(self, supplied_data, block_number: int = "latest") -> dict:
        pass

    def decode_lp_token_info(self, supplied_data, decoded_data, block_number: int = "latest") -> dict:
        pass

    # Get lp list
    def get_all_supported_lp_token(self, limit: int = 10, supplied_data: dict = None) -> dict:
        pass

    def decode_all_supported_lp_token(self, limit: int = 10, decoded_data: dict = None,
                                      supplied_data: dict = None) -> dict:
        pass

    # Get farming supported lp list
    def get_farming_supported_lp_token(self, limit: int = 10) -> dict:
        pass

    def decode_farming_supported_lp_token(self, decoded_data) -> dict:
        pass

    def get_important_lp_token(self, supplied_data, block_number) -> dict:
        pass

    def decode_important_lp_token(self, supplied_data, decoded_data) -> dict:
        pass

    # Lp token liquidity
    def get_balance_of_token_function_info(self, supplied_data, block_number: int = "latest") -> dict:
        pass

    def decode_balance_of_token_function_info(
            self, supplied_data, decoded_data, block_number: int = "latest") -> dict:
        pass

    # User information
    def get_user_info_function(self, wallet: str, supplied_data: dict, stake: bool = True,
                               block_number: int = "latest") -> dict:
        pass

    def decode_user_info_function(self, wallet: str, supplied_data: dict, user_data: dict, stake: bool = True,
                                  block_number: int = "latest") -> dict:
        pass

    def get_user_token_amount_function(self, wallet: str, supplied_data: dict, block_number: int = 'latest') -> dict:
        pass

    def decode_user_token_amount_function(self, user: str, supplied_data: dict, decoded_data: dict = None,
                                          block_number: int = "latest") -> dict:
        pass

    def get_all_nft_token_of_user_function(
            self, wallet: str, block_number: int = "latest") -> dict:
        pass

    def decode_all_nft_token_of_user_function(
            self, decode_data: dict) -> dict:
        pass

    # User reward
    def get_rewards_balance_function_info(self, wallet, supplied_data, block_number: int = "latest") -> dict:
        pass

    def calculate_rewards_balance(
            self, wallet: str, supplied_data: dict, decoded_data: dict, block_number: int = "latest") -> dict:
        pass

    @staticmethod
    def get_token_list() -> list:
        return []



class Lending:
    # service
    aave_v2 = "aave-v2"
    aave_v3 = "aave-v3"
    radiant_v2 = "radiant-v2"
    compound = "compound"
    flux = "flux-finance"
    iron_bank = "iron-bank"
    trava = "trava-finance"
    valas = "valas-finance"
    geist = "geist-finance"
    cream = "cream-lending"
    venus = "venus"
    liqee = "liqee"
    strike = "strike"
    uwu = "uwu-lend"
    onyx = "onyx-protocol"
    granary = "granary-finance"
    wepiggy = "wepiggy"
    morpho_compound = 'morpho-compound'
    morpho_aave_v2 = 'morpho-aave'
    morpho_aave_v3 = 'morpho-aavev3'
    spark = 'spark'
    ape_swap = "apeswap-lending"
    silo = 'silo-finance'
    compound_v3 = 'compound-v3'
    justlend = "justlend"
    all = [strike, aave_v2, aave_v3, radiant_v2, compound, flux, onyx, granary,
           iron_bank, trava, valas, geist, cream, venus, liqee, strike, uwu, wepiggy,
           morpho_compound, morpho_aave_v2, morpho_aave_v3, spark, ape_swap, silo, compound_v3,
           justlend
           ]



import logging
import time

from web3 import Web3

from defi_services.abis.lending.aave_v2_and_forks.aave_v2_incentives_abi import AAVE_V2_INCENTIVES_ABI
from defi_services.abis.lending.aave_v2_and_forks.lending_pool_abi import LENDING_POOL_ABI
from defi_services.abis.lending.aave_v2_and_forks.oracle_abi import ORACLE_ABI
from defi_services.abis.lending.morpho.morpho_aave_v2_comptroller_abi import MORPHO_AAVE_V2_COMPTROLLER
from defi_services.abis.lending.morpho.morpho_aave_v2_lens_abi import MORPHO_AAVE_V2_LENS_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.lending_constant import Lending
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.lending.aave_v2_services import AaveInfo
from defi_services.services.lending.lending_info.ethereum.morpho_aave_v2_eth import MORPHO_AAVE_V2_ETH
from defi_services.services.lending.morpho_compound_services import MorphoCompoundStateService
from defi_services.utils.apy import apr_to_apy

logger = logging.getLogger("Compound Lending Pool State Service")


class MorphoAaveV2Info:
    mapping = {
        Chain.ethereum: MORPHO_AAVE_V2_ETH
    }


class MorphoAaveV2StateService(MorphoCompoundStateService):
    def __init__(self, state_service: StateQuerier, chain_id: str = "0x1"):
        super().__init__(state_service, chain_id)
        self.name = f"{chain_id}_{Lending.morpho_aave_v2}"
        self.chain_id = chain_id
        self.aave_info = AaveInfo.mapping.get(chain_id)
        self.pool_info = MorphoAaveV2Info.mapping.get(chain_id)
        self.state_service = state_service
        self.lending_abi = LENDING_POOL_ABI
        self.incentive_abi = AAVE_V2_INCENTIVES_ABI
        self.oracle_abi = ORACLE_ABI
        self.lens_abi = MORPHO_AAVE_V2_LENS_ABI
        self.comptroller_abi = MORPHO_AAVE_V2_COMPTROLLER
        self.market_key = 'tToken'

    # BASIC FUNCTIONS
    def get_service_info(self):
        info = {
            Lending.morpho_aave_v2: {
                "chain_id": self.chain_id,
                "type": "lending",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_dapp_asset_info(
            self,
            block_number: int = "latest"):
        begin = time.time()
        _w3 = self.state_service.get_w3()
        pool_address = Web3.to_checksum_address(self.aave_info['address'])
        contract = _w3.eth.contract(address=pool_address, abi=self.lending_abi)
        comptroller_contract = _w3.eth.contract(
            address=_w3.to_checksum_address(self.pool_info.get("comptrollerAddress")), abi=self.comptroller_abi)
        markets = comptroller_contract.functions.getMarketsCreated().call(block_identifier=block_number)
        reserves_list = contract.functions.getReservesList().call(block_identifier=block_number)
        reserves_info = {}
        for token in reserves_list:
            value = contract.functions.getReserveData(token).call(block_identifier=block_number)
            key = token.lower()
            ttoken = value[7]
            if ttoken in markets:
                reserves_info[key] = {}
                reserves_info[key]["tToken"] = ttoken.lower()
                reserves_info[key]["dToken"] = value[9].lower()
                reserves_info[key]["sdToken"] = value[8].lower()

                risk_param = bin(value[0][0])[2:]
                reserves_info[key]["loanToValue"] = int(risk_param[-15:], 2) / 10 ** 4
                reserves_info[key]["liquidationThreshold"] = int(risk_param[-31:-16], 2) / 10 ** 4
        logger.info(f"Get reserves information in {time.time() - begin}s")
        return reserves_info

    # CALCULATE APY LENDING POOL
    def get_apy_lending_pool_function_info(
            self,
            reserves_info: dict,
            block_number: int = "latest"
    ):
        rpc_calls = {}
        for token_address, value in reserves_info.items():
            decimals_key = f"decimals_{token_address}_{block_number}".lower()
            rpc_calls[decimals_key] = self.state_service.get_function_info(
                token_address, ERC20_ABI, "decimals", block_number=block_number)

            a_token = value['tToken']
            market_data_query_id = f"getMainMarketData_{self.name}_{a_token}_{block_number}".lower()
            rpc_calls[market_data_query_id] = self.get_lens_function_info(
                'getMainMarketData', [a_token], block_number=block_number)

            reserve_key = f"getReserveData_{self.name}_{token_address}_{block_number}".lower()
            rpc_calls[reserve_key] = self.get_function_lending_pool_info("getReserveData", [token_address])

        return rpc_calls

    def get_reserve_tokens_metadata(
            self,
            decoded_data: dict,
            reserves_info: dict,
            block_number: int = "latest"
    ):
        reserve_tokens_info = []
        for token_address, reserve_info in reserves_info.items():
            get_reserve_data_call_id = f'getReserveData_{self.name}_{token_address}_{block_number}'.lower()
            reserve_data = decoded_data.get(get_reserve_data_call_id)

            a_token = reserve_data[7].lower()
            market_data_query_id = f"getMainMarketData_{self.name}_{a_token}_{block_number}".lower()
            market_data = decoded_data.get(market_data_query_id)

            decimals_call_id = f"decimals_{token_address}_{block_number}".lower()

            reserve_tokens_info.append({
                'underlying': token_address,
                'underlying_decimals': decoded_data.get(decimals_call_id),
                'supply_apy': reserve_data[3],
                'borrow_apy': reserve_data[4],
                'p2p_supply': market_data[2],
                'p2p_borrow': market_data[3],
                'pool_supply': market_data[4],
                'pool_borrow': market_data[5]
            })

        return reserve_tokens_info

    def calculate_apy_lending_pool_function_call(
            self,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict,
            pool_token_price: float,
            pool_decimals: int = 18,
            block_number: int = 'latest',
    ):
        reserve_tokens_info = self.get_reserve_tokens_metadata(decoded_data, reserves_info, block_number)

        data = {}
        for token_info in reserve_tokens_info:
            underlying_token = token_info['underlying']
            data[underlying_token] = self._calculate_interest_rates(
                token_info, pool_decimals=pool_decimals,
                apx_block_speed_in_seconds=0
            )

        return {self.pool_info.get("comptrollerAddress"): data}

    @classmethod
    def _calculate_interest_rates(cls, token_info: dict, pool_decimals: int, apx_block_speed_in_seconds: int):
        decimals = token_info['underlying_decimals']
        total_supply = float(token_info["p2p_supply"]) / 10 ** decimals + float(token_info["pool_supply"]) / 10 ** decimals
        total_borrow = float(token_info["p2p_borrow"]) / 10 ** decimals + float(token_info["pool_borrow"]) / 10 ** decimals

        supply_apr = float(token_info['supply_apy']) / 10 ** 27
        supply_apy = apr_to_apy(supply_apr)
        borrow_apr = float(token_info['borrow_apy']) / 10 ** 27
        borrow_apy = apr_to_apy(borrow_apr)

        return {
            'deposit_apy': supply_apy,
            'borrow_apy': borrow_apy,
            'total_deposit': total_supply,
            'total_borrow': total_borrow
        }

    # REWARDS BALANCE
    def get_rewards_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict = None,
            block_number: int = "latest",
    ):
        return {}

    def calculate_rewards_balance(self, wallet: str, reserves_info: dict, decoded_data: dict, block_number: int = "latest"):
        return {}

    def get_function_lending_pool_info(self, fn_name: str, fn_paras=None, block_number: int = 'latest'):
        return self.state_service.get_function_info(
            self.aave_info['address'], self.lending_abi, fn_name, fn_paras, block_number
        )

    def get_lens_function_info(self, fn_name: str, fn_paras: list, block_number: int = "latest"):
        return self.state_service.get_function_info(
            self.pool_info['lensAddress'], self.lens_abi, fn_name, fn_paras, block_number
        )



import logging
import time

from web3 import Web3

from defi_services.abis.lending.aave_v2_and_forks.aave_v2_incentives_abi import AAVE_V2_INCENTIVES_ABI
from defi_services.abis.lending.aave_v2_and_forks.oracle_abi import ORACLE_ABI
from defi_services.abis.lending.aave.aave_v3.aave_v3_lending_pool_abi import AAVE_V3_LENDING_POOL_ABI
from defi_services.abis.lending.morpho.morpho_aave_v3_comptroller_abi import MORPHO_AAVE_V3_COMPTROLLER_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.lending_constant import Lending
from defi_services.constants.token_constant import Token
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.lending.aave_v3_services import AaveV3Info
from defi_services.services.lending.lending_info.ethereum.morpho_aave_v3_eth import MORPHO_AAVE_V3_ETH
from defi_services.services.lending.morpho_compound_services import MorphoCompoundStateService

logger = logging.getLogger("Compound Lending Pool State Service")


class MorphoAaveV3Info:
    mapping = {
        Chain.ethereum: MORPHO_AAVE_V3_ETH
    }


class MorphoAaveV3StateService(MorphoCompoundStateService):
    def __init__(self, state_service: StateQuerier, chain_id: str = "0x1"):
        super().__init__(state_service, chain_id)
        self.name = f"{chain_id}_{Lending.morpho_aave_v3}"
        self.chain_id = chain_id
        self.aave_info = AaveV3Info.mapping.get(chain_id)
        self.pool_info = MorphoAaveV3Info.mapping.get(chain_id)
        self.state_service = state_service
        self.lending_abi = AAVE_V3_LENDING_POOL_ABI
        self.incentive_abi = AAVE_V2_INCENTIVES_ABI
        self.oracle_abi = ORACLE_ABI
        self.comptroller_abi = MORPHO_AAVE_V3_COMPTROLLER_ABI
        self.market_key = 'tToken'

    # BASIC FUNCTIONS
    def get_service_info(self):
        info = {
            Lending.morpho_aave_v3: {
                "chain_id": self.chain_id,
                "type": "lending",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_dapp_asset_info(
            self,
            block_number: int = "latest"):
        begin = time.time()
        _w3 = self.state_service.get_w3()
        pool_address = Web3.to_checksum_address(self.aave_info['address'])
        contract = _w3.eth.contract(address=pool_address, abi=self.lending_abi)
        comptroller_contract = _w3.eth.contract(
            address=_w3.to_checksum_address(self.pool_info.get("comptrollerAddress")), abi=self.comptroller_abi)
        markets = comptroller_contract.functions.marketsCreated().call(block_identifier=block_number)
        markets = [i.lower() for i in markets]
        reserves_list = contract.functions.getReservesList().call(block_identifier=block_number)
        reserves_info = {}
        for token in reserves_list:
            value = contract.functions.getReserveData(token).call(block_identifier=block_number)
            key = token.lower()
            if key in markets:
                reserves_info[key] = {}
                reserves_info[key]["tToken"] = value[8].lower()
                reserves_info[key]["sdToken"] = value[9].lower()
                reserves_info[key]["dToken"] = value[10].lower()

                risk_param = bin(value[0][0])[2:]
                reserves_info[key]["loanToValue"] = int(risk_param[-15:], 2) / 10 ** 4
                reserves_info[key]["liquidationThreshold"] = int(risk_param[-31:-16], 2) / 10 ** 4
        logger.info(f"Get reserves information in {time.time() - begin}s")
        return reserves_info

    # CALCULATE APY LENDING POOL
    def get_apy_lending_pool_function_info(
            self,
            reserves_info: dict,
            block_number: int = "latest"
    ):
        rpc_calls = {}
        # TODO: get rpc call for Morpho Aave V2 total supply, total borrow, and APY
        return rpc_calls

    def calculate_apy_lending_pool_function_call(
            self,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict,
            pool_token_price: float,
            pool_decimals: int = 18,
            block_number: int = 'latest',
    ):
        data = {}
        # TODO: get calculate Morpho Aave V2 total supply, total borrow, and APY
        return data

    # REWARDS BALANCE
    def get_rewards_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict = None,
            block_number: int = "latest",
    ):
        return {}

    def calculate_rewards_balance(
            self, wallet: str, reserves_info: dict, decoded_data: dict, block_number: int = "latest"):
        return {}

    # WALLET DEPOSIT BORROW BALANCE
    def get_wallet_deposit_borrow_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict,
            block_number: int = "latest"
    ):

        rpc_calls = {}
        for token, value in reserves_info.items():
            underlying = token
            ctoken = value.get(self.market_key)
            if token == Token.native_token:
                underlying = Token.wrapped_token.get(self.chain_id)
            underlying_borrow_key = f"borrowBalance_{self.name}_{ctoken}_{wallet}_{block_number}".lower()
            underlying_balance_key = f"collateralBalance_{self.name}_{ctoken}_{wallet}_{block_number}".lower()
            underlying_decimals_key = f"decimals_{underlying}_{block_number}".lower()
            rpc_calls[underlying_borrow_key] = self.get_comptroller_function_info(
                "borrowBalance", [token, wallet], block_number)
            rpc_calls[underlying_balance_key] = self.get_comptroller_function_info(
                "collateralBalance", [token, wallet], block_number)
            rpc_calls[underlying_decimals_key] = self.state_service.get_function_info(
                underlying, ERC20_ABI, "decimals", [], block_number
            )

        return rpc_calls

    def calculate_wallet_deposit_borrow_balance(
            self,
            wallet: str,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict = None,
            pool_decimals: int = 18,
            block_number: int = "latest"
    ):
        if token_prices is None:
            token_prices = {}
        result = {}
        pool_address = self.pool_info.get("comptrollerAddress")
        for token, value in reserves_info.items():
            data = {}
            underlying = token
            ctoken = value.get(self.market_key)
            if token == Token.native_token:
                underlying = Token.wrapped_token.get(self.chain_id)
            get_total_deposit_id = f"collateralBalance_{self.name}_{ctoken}_{wallet}_{block_number}".lower()
            get_total_borrow_id = f"borrowBalance_{self.name}_{ctoken}_{wallet}_{block_number}".lower()
            get_decimals_id = f"decimals_{underlying}_{block_number}".lower()
            decimals = decoded_data[get_decimals_id]
            deposit_amount = decoded_data[get_total_deposit_id] / 10 ** decimals
            borrow_amount = decoded_data[get_total_borrow_id] / 10 ** decimals
            data[token] = {
                "borrow_amount": borrow_amount,
                "deposit_amount": deposit_amount,
                "is_collateral": True if value.get('liquidationThreshold') > 0 else False
            }
            if token_prices:
                token_price = token_prices.get(underlying)
            else:
                token_price = None
            if token_price is not None:
                deposit_amount_in_usd = deposit_amount * token_price
                borrow_amount_in_usd = borrow_amount * token_price
                data[token]['borrow_amount_in_usd'] += borrow_amount_in_usd
                data[token]['deposit_amount_in_usd'] += deposit_amount_in_usd
            result.update(data)
        return {pool_address.lower(): result}

    def get_function_lending_pool_info(self, fn_name: str, fn_paras=None, block_number: int = 'latest'):
        return self.state_service.get_function_info(
            self.aave_info['address'], self.lending_abi, fn_name, fn_paras, block_number
        )



import json

LIQEE_LENDING_DATA_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_controller",
        "type": "address"
      },
      {
        "internalType": "contract IiToken",
        "name": "_priceToken",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "NewOwner",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "oldPendingOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newPendingOwner",
        "type": "address"
      }
    ],
    "name": "NewPendingOwner",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "_acceptOwner",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address payable",
        "name": "newPendingOwner",
        "type": "address"
      }
    ],
    "name": "_setPendingOwner",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "amounts",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "blocksPerYear",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "calcAccountEquity",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_safeMaxFactor",
        "type": "uint256"
      }
    ],
    "name": "canAccountRemoveFromCollateral",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "controller",
    "outputs": [
      {
        "internalType": "contract IController",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "decimals",
    "outputs": [
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountAvailable",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_safeMaxFactor",
        "type": "uint256"
      }
    ],
    "name": "getAccountBorrowData",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_safeMaxFactor",
        "type": "uint256"
      }
    ],
    "name": "getAccountBorrowInfo",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountBorrowStatus",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountBorrowTokens",
    "outputs": [
      {
        "internalType": "contract IiToken[]",
        "name": "",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      },
      {
        "internalType": "uint8[]",
        "name": "",
        "type": "uint8[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountBorrowValue",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_borrowValue",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountMSDTokens",
    "outputs": [
      {
        "internalType": "contract IiToken[]",
        "name": "",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      },
      {
        "internalType": "uint8[]",
        "name": "",
        "type": "uint8[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountRewardAmount",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_safeMaxFactor",
        "type": "uint256"
      }
    ],
    "name": "getAccountSupplyData",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_safeMaxFactor",
        "type": "uint256"
      }
    ],
    "name": "getAccountSupplyInfo",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_assetPrice",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "_asCollateral",
        "type": "bool"
      },
      {
        "internalType": "bool",
        "name": "_executed",
        "type": "bool"
      },
      {
        "internalType": "bool",
        "name": "_accountAvailable",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountSupplyTokens",
    "outputs": [
      {
        "internalType": "contract IiToken[]",
        "name": "",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      },
      {
        "internalType": "uint8[]",
        "name": "",
        "type": "uint8[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountTokens",
    "outputs": [
      {
        "internalType": "contract IiToken[]",
        "name": "_supplyTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "_supplyAmounts",
        "type": "uint256[]"
      },
      {
        "internalType": "uint8[]",
        "name": "_supplyDecimals",
        "type": "uint8[]"
      },
      {
        "internalType": "contract IiToken[]",
        "name": "_borrowTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "_borrowAmounts",
        "type": "uint256[]"
      },
      {
        "internalType": "uint8[]",
        "name": "_borrowDecimals",
        "type": "uint8[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getAccountTotalValue",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "getAssetUSDPrice",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_account",
        "type": "address"
      }
    ],
    "name": "getBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "getBorrowTokenData",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_borrower",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_liquidator",
        "type": "address"
      },
      {
        "internalType": "contract IiToken",
        "name": "_assetBorrowed",
        "type": "address"
      },
      {
        "internalType": "contract IiToken",
        "name": "_assetCollateral",
        "type": "address"
      }
    ],
    "name": "getLiquidationInfo",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "getSupplyTokenData",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_controller",
        "type": "address"
      },
      {
        "internalType": "contract IiToken",
        "name": "_priceToken",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address payable",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pendingOwner",
    "outputs": [
      {
        "internalType": "address payable",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "priceToken",
    "outputs": [
      {
        "internalType": "contract IiToken",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IController",
        "name": "_newController",
        "type": "address"
      }
    ],
    "name": "setController",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract IiToken",
        "name": "_newAsset",
        "type": "address"
      }
    ],
    "name": "setPriceToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "tokens",
    "outputs": [
      {
        "internalType": "contract IiToken",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')



import json

REWARDS_CONTROLLER = json.loads("""
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "emissionManager",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "assetIndex",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "userIndex",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardsAccrued",
        "type": "uint256"
      }
    ],
    "name": "Accrued",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldEmission",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newEmission",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldDistributionEnd",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newDistributionEnd",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "assetIndex",
        "type": "uint256"
      }
    ],
    "name": "AssetConfigUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "claimer",
        "type": "address"
      }
    ],
    "name": "ClaimerSet",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "oldEmissionManager",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newEmissionManager",
        "type": "address"
      }
    ],
    "name": "EmissionManagerUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "rewardOracle",
        "type": "address"
      }
    ],
    "name": "RewardOracleUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "claimer",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "RewardsClaimed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "transferStrategy",
        "type": "address"
      }
    ],
    "name": "TransferStrategyInstalled",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "REVISION",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      }
    ],
    "name": "claimAllRewards",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "rewardsList",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "claimedAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      }
    ],
    "name": "claimAllRewardsOnBehalf",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "rewardsList",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "claimedAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      }
    ],
    "name": "claimAllRewardsToSelf",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "rewardsList",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "claimedAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "claimRewards",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "claimRewardsOnBehalf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "claimRewardsToSelf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "uint88",
            "name": "emissionPerSecond",
            "type": "uint88"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint32",
            "name": "distributionEnd",
            "type": "uint32"
          },
          {
            "internalType": "address",
            "name": "asset",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "reward",
            "type": "address"
          },
          {
            "internalType": "contract ITransferStrategyBase",
            "name": "transferStrategy",
            "type": "address"
          },
          {
            "internalType": "contract IEACAggregatorProxy",
            "name": "rewardOracle",
            "type": "address"
          }
        ],
        "internalType": "struct RewardsDataTypes.RewardsConfigInput[]",
        "name": "config",
        "type": "tuple[]"
      }
    ],
    "name": "configureAssets",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "getAllUserRewards",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "rewardsList",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "unclaimedAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getAssetDecimals",
    "outputs": [
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      }
    ],
    "name": "getClaimer",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getDistributionEnd",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getEmissionManager",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getRewardOracle",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getRewardsByAsset",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getRewardsData",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getRewardsList",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getTransferStrategy",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getUserAccruedRewards",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getUserAssetIndex",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      }
    ],
    "name": "getUserRewards",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "totalSupply",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "userBalance",
        "type": "uint256"
      }
    ],
    "name": "handleAction",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "emissionManager",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "caller",
        "type": "address"
      }
    ],
    "name": "setClaimer",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "internalType": "uint32",
        "name": "newDistributionEnd",
        "type": "uint32"
      }
    ],
    "name": "setDistributionEnd",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "emissionManager",
        "type": "address"
      }
    ],
    "name": "setEmissionManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "address[]",
        "name": "rewards",
        "type": "address[]"
      },
      {
        "internalType": "uint88[]",
        "name": "newEmissionsPerSecond",
        "type": "uint88[]"
      }
    ],
    "name": "setEmissionPerSecond",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "internalType": "contract IEACAggregatorProxy",
        "name": "rewardOracle",
        "type": "address"
      }
    ],
    "name": "setRewardOracle",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "reward",
        "type": "address"
      },
      {
        "internalType": "contract ITransferStrategyBase",
        "name": "transferStrategy",
        "type": "address"
      }
    ],
    "name": "setTransferStrategy",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
""")
        


# import json
#
# from defi_services.constants.entities.dex_constant import Dex
# from defi_services.constants.query_constant import Query
# from defi_services.jobs.processors.state_processor import StateProcessor
# from src.defi_services.databases.mongodb_klg import MongoDB
#
# provider_url = {
#     "0x38": 'https://bsc-dataseed3.binance.org/',
#     '0x1': "https://rpc.ankr.com/eth",
#     '0xfa': "https://rpc.ankr.com/fantom",
#     '0xa': "https://optimism.llamarpc.com",
#     '0xa4b1': "https://rpc.ankr.com/arbitrum",
#     '0xa86a': "https://rpc.ankr.com/avalanche",
#     '0x89': "https://rpc.ankr.com/polygon"}
#
#
# # TEST LP TOKEN INFO ##
# def test_lp_token_processor(_dex_protocol):
#     queries = [
#         {
#             'query_id': f'{_dex_protocol}_lptokeninfo',
#             "entity_id": _dex_protocol,
#             'query_type': 'lp_token_list',
#             'number_lp': 10
#         }]
#
#     lp_token_list = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
#     lp_token_inf = get_lp_token_info(lp_token_list, _dex_protocol)
#     lp_token_token_info = get_stake_token_pair(lp_token_inf, _dex_protocol)
#     with open('lp_token_info1.json', 'w') as f:
#         json.dump(lp_token_token_info, f)
#
#
# def get_lp_token_info(data, _dex_protocol):
#     queries = [{
#         'query_id': f'{_dex_protocol}_lptokeninfo',
#         "entity_id": _dex_protocol,
#         'query_type': 'lp_token_info',
#         'supplied_data': {
#             'lp_token_info': data[0]['lp_token_list']}
#     }]
#     res = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
#     return res
#
#
# def get_stake_token_pair(data, _dex_protocol):
#     lp_token_info = data[0]['lp_token_info']
#     token_list = []
#     for lp_token, info in lp_token_info.items():
#         token0 = info.get("token0", None)
#         token1 = info.get("token1", None)
#         if token0 and token0 not in token_list:
#             token_list.append(token0)
#         if token1 and token1 not in token_list:
#             token_list.append(token1)
#     token_info = get_token_info(token_list)
#     queries = [{
#         'query_id': f'{_dex_protocol}_lptokeninfo',
#         "entity_id": _dex_protocol,
#         'query_type': 'token_pair_balance',
#         'supplied_data': {
#             'lp_token_info': lp_token_info,
#             'token_info': token_info
#         }
#     }]
#     res = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
#     return res
#
#
# def get_token_info(token_list):
#     token_info = {}
#     mongo_klg = MongoDB()
#     for token in token_list:
#         token_data = mongo_klg.get_smart_contract(chain_id, token)
#         if token_data:
#             price = token_data.get("price", 0)
#             decimal = token_data.get("decimals", 18)
#             token_info[token] = {
#                 'price': price,
#                 'decimals': decimal}
#     return token_info
#
#
# # TEST USER INFO ##
# def test_user_info_processor(_dex_protocol, list_pool: list = None):
#     with open('lp_token_info1.json', 'r') as f:
#         data = json.loads(f.read())
#     lp_token_info = data[0]['token_pair_balance']
#     token_list = []
#     for lp_token, info in lp_token_info.items():
#         token0 = info.get("token0", None)
#         token1 = info.get("token1", None)
#         if token0 and token0 not in token_list:
#             token_list.append(token0)
#         if token1 and token1 not in token_list:
#             token_list.append(token1)
#
#     if list_pool:
#         farm = {}
#         pools = []
#         for pool in list_pool:
#             if pool in lp_token_info:
#                 farm.update(lp_token_info[pool])
#             else:
#                 pools.append(pool)
#         queries = [
#             {
#                 'query_id': f'{_dex_protocol}_userinfo',
#                 "entity_id": _dex_protocol,
#                 'query_type': 'dex_user_info',
#                 'supplied_data': {
#                     'lp_token_info': farm
#                 },
#                 'stake': True
#             },
#             {
#                 'query_id': f'{_dex_protocol}_userinfo',
#                 "entity_id": _dex_protocol,
#                 'query_type': 'dex_user_info',
#                 'supplied_data': {
#                     'lp_token_info': pools,
#                 },
#                 'stake': False
#             },
#             {
#                 'query_id': f'{_dex_protocol}_userinfo',
#                 "entity_id": _dex_protocol,
#                 'query_type': 'protocol_reward',
#                 'supplied_data': {
#                     'lp_token_info': farm}
#             }
#         ]
#     else:
#         queries = [
#             {
#                 'query_id': f'{Query.dex_user_info}_{Dex.pancake_v2}',
#                 "entity_id": Dex.pancake_v2,
#                 'query_type': Query.dex_user_info,
#                 'supplied_data': {
#                     'lp_token_info': lp_token_info
#                 },
#                 'stake': True
#             },
#             {
#                 'query_id': f'{Query.protocol_reward}_{Dex.pancake_v2}',
#                 "entity_id": Dex.pancake_v2,
#                 'query_type': Query.protocol_reward,
#                 'supplied_data': {
#                     'lp_token_info': lp_token_info
#                 }
#             }
#         ]
#
#     res = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
#     print(res)
#
#     # For another version
#     # with open('user.json', 'w') as f:
#     #     json.dump(res, f)
#
#
# if __name__ == "__main__":
#     chain_id = '0x38'
#     wallet = "0x1b2a2f6ed4a1401e8c73b4c2b6172455ce2f78e8"
#     job = StateProcessor(provider_url[chain_id], chain_id)
#     dex_protocol = 'pancakeswap'
#     test_lp_token_processor(dex_protocol,)
#     # test_user_info_processor(dex_protocol)



import json

from defi_services.constants.entities.dex_constant import Dex
from defi_services.constants.query_constant import Query
from defi_services.jobs.processors.state_processor import StateProcessor

chain_id = '0x38'

provider_url = {
    "0x38": 'https://bsc-dataseed3.binance.org/',
    '0x1': "https://rpc.ankr.com/eth",
    '0xfa': "https://rpc.ankr.com/fantom",
    '0xa': "https://optimism.llamarpc.com",
    '0xa4b1': "https://rpc.ankr.com/arbitrum",
    '0xa86a': "https://rpc.ankr.com/avalanche",
    '0x89': "https://rpc.ankr.com/polygon"}


def test_v3():
    wallet = "0x63B59cb9F03Bc57DF16d7a45423cE8148B4818D9"
    job = StateProcessor(provider_url[chain_id], chain_id)
    queries = [
        {
            'query_id': f'{Query.lp_token_list}_{Dex.pancake_v3}',
            "entity_id": Dex.pancake_v3,
            'query_type': Query.lp_token_list
        }]

    res = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
    print(res)

    # For another version
    # with open('../test/result.json', 'w') as f:
    #     json.dump(res, f)


def test_user_v3():
    wallet = "0xc863ffec450e71baa242fb2efbb20f6419df0cc6"
    job = StateProcessor(provider_url[chain_id], chain_id)

    queries = [
        {
            'query_id': f'{Query.dex_user_nft}_{Dex.pancake_v3}',
            "entity_id": Dex.pancake_v3,
            'query_type': Query.dex_user_nft,
        }
    ]

    res = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
    print(res)

    # For another version
    # with open('../test/user.json', 'w') as f:
    #     json.dump(res, f)


def test_user_info_v3():
    wallet = "0xc863ffec450e71baa242fb2efbb20f6419df0cc6"
    job = StateProcessor(provider_url[chain_id], chain_id)
    with open('user.json', 'r') as f:
        data = json.loads(f.read())
    supplied_data = data[0]['dex_user_nft']

    queries = [
        {
            'query_id': f'{Query.dex_user_info}_{Dex.pancake_v3}',
            "entity_id": Dex.pancake_v3,
            'query_type': Query.dex_user_info,
            'supplied_data': supplied_data
        }
    ]

    res = job.run(wallet, queries, batch_size=100, max_workers=8, ignore_error=True)
    print(res)

    # For another version
    # with open('../test/user.json', 'w') as f:
    #     json.dump(res, f)



import json

SILO_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "contract ISiloRepository",
        "name": "_repository",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_siloAsset",
        "type": "address"
      },
      {
        "internalType": "uint128",
        "name": "_version",
        "type": "uint128"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "AssetDoesNotExist",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BorrowNotPossible",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "DepositNotPossible",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "DepositsExceedLimit",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "DifferentArrayLength",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidRepository",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidSiloVersion",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "LiquidationReentrancyCall",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "MaximumLTVReached",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NotEnoughDeposits",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NotEnoughLiquidity",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NotSolvent",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "OnlyRouter",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "Paused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TokenIsNotAContract",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnexpectedEmptyReturn",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UnsupportedLTVType",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "UserIsZero",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroAssets",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroShares",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "enum IBaseSilo.AssetStatus",
        "name": "status",
        "type": "uint8"
      }
    ],
    "name": "AssetStatusUpdate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "Borrow",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "depositor",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "collateralOnly",
        "type": "bool"
      }
    ],
    "name": "Deposit",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "shareAmountRepaid",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "seizedCollateral",
        "type": "uint256"
      }
    ],
    "name": "Liquidate",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "Repay",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "depositor",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "collateralOnly",
        "type": "bool"
      }
    ],
    "name": "Withdraw",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "VERSION",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "accrueInterest",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "interest",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "assetStorage",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract IShareToken",
            "name": "collateralToken",
            "type": "address"
          },
          {
            "internalType": "contract IShareToken",
            "name": "collateralOnlyToken",
            "type": "address"
          },
          {
            "internalType": "contract IShareToken",
            "name": "debtToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "totalDeposits",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "collateralOnlyDeposits",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrowAmount",
            "type": "uint256"
          }
        ],
        "internalType": "struct IBaseSilo.AssetStorage",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "borrow",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "debtAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "debtShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_borrower",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_receiver",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "borrowFor",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "debtAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "debtShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_borrower",
        "type": "address"
      }
    ],
    "name": "borrowPossible",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "_collateralOnly",
        "type": "bool"
      }
    ],
    "name": "deposit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "collateralAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "collateralShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_depositor",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "_collateralOnly",
        "type": "bool"
      }
    ],
    "name": "depositFor",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "collateralAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "collateralShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_depositor",
        "type": "address"
      }
    ],
    "name": "depositPossible",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "_users",
        "type": "address[]"
      },
      {
        "internalType": "bytes",
        "name": "_flashReceiverData",
        "type": "bytes"
      }
    ],
    "name": "flashLiquidate",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "internalType": "uint256[][]",
        "name": "receivedCollaterals",
        "type": "uint256[][]"
      },
      {
        "internalType": "uint256[][]",
        "name": "shareAmountsToRepay",
        "type": "uint256[][]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAssets",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAssetsWithState",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "assets",
        "type": "address[]"
      },
      {
        "components": [
          {
            "internalType": "contract IShareToken",
            "name": "collateralToken",
            "type": "address"
          },
          {
            "internalType": "contract IShareToken",
            "name": "collateralOnlyToken",
            "type": "address"
          },
          {
            "internalType": "contract IShareToken",
            "name": "debtToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "totalDeposits",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "collateralOnlyDeposits",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrowAmount",
            "type": "uint256"
          }
        ],
        "internalType": "struct IBaseSilo.AssetStorage[]",
        "name": "assetsStorage",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "harvestProtocolFees",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "harvestedAmounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "initAssetsTokens",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "interestData",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "harvestedProtocolFees",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "protocolFees",
            "type": "uint256"
          },
          {
            "internalType": "uint64",
            "name": "interestRateTimestamp",
            "type": "uint64"
          },
          {
            "internalType": "enum IBaseSilo.AssetStatus",
            "name": "status",
            "type": "uint8"
          }
        ],
        "internalType": "struct IBaseSilo.AssetInterestData",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      }
    ],
    "name": "isSolvent",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "liquidity",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "repay",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "repaidAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "repaidShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_borrower",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "repayFor",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "repaidAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "repaidShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "siloAsset",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "siloRepository",
    "outputs": [
      {
        "internalType": "contract ISiloRepository",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "syncBridgeAssets",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      }
    ],
    "name": "utilizationData",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "totalDeposits",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrowAmount",
            "type": "uint256"
          },
          {
            "internalType": "uint64",
            "name": "interestRateTimestamp",
            "type": "uint64"
          }
        ],
        "internalType": "struct IBaseSilo.UtilizationData",
        "name": "data",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "_collateralOnly",
        "type": "bool"
      }
    ],
    "name": "withdraw",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "withdrawnAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "withdrawnShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_asset",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_depositor",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_receiver",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "_collateralOnly",
        "type": "bool"
      }
    ],
    "name": "withdrawFor",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "withdrawnAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "withdrawnShare",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


PANCAKESWAP_V2_FACTORY_ABI = [
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_feeToSetter",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "token0",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "token1",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "pair",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "", "type": "uint256"
            }
        ],
        "name": "PairCreated",
        "type": "event"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "INIT_CODE_PAIR_HASH",
        "outputs": [
            {
                "internalType": "bytes32",
                "name": "",
                "type": "bytes32"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "name": "allPairs",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "allPairsLength",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "tokenA",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "tokenB",
                "type": "address"
            }
        ],
        "name": "createPair",
        "outputs": [
            {
                "internalType": "address",
                "name": "pair",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "feeTo",
        "outputs": [
            {"internalType": "address",
             "name": "",
             "type": "address"
             }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "feeToSetter",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "getPair",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "_feeTo",
                "type": "address"
            }
        ],
        "name": "setFeeTo",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "_feeToSetter",
                "type": "address"
            }
        ],
        "name": "setFeeToSetter",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]




import json

REWARDSDISTRIBUTORIMPL = json.loads("""
[
  {
    "inputs": [],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "loopsLimit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "requiredLoops",
        "type": "uint256"
      }
    ],
    "name": "MaxLoopsLimitExceeded",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "calledContract",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "methodSignature",
        "type": "string"
      }
    ],
    "name": "Unauthorized",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint32",
        "name": "newBlock",
        "type": "uint32"
      }
    ],
    "name": "BorrowLastRewardingBlockUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "contributor",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newSpeed",
        "type": "uint256"
      }
    ],
    "name": "ContributorRewardTokenSpeedUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "contributor",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardAccrued",
        "type": "uint256"
      }
    ],
    "name": "ContributorRewardsUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardTokenDelta",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardTokenTotal",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardTokenBorrowIndex",
        "type": "uint256"
      }
    ],
    "name": "DistributedBorrowerRewardToken",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "supplier",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardTokenDelta",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardTokenTotal",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "rewardTokenSupplyIndex",
        "type": "uint256"
      }
    ],
    "name": "DistributedSupplierRewardToken",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "MarketInitialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "oldMaxLoopsLimit",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newmaxLoopsLimit",
        "type": "uint256"
      }
    ],
    "name": "MaxLoopsLimitUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "oldAccessControlManager",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "newAccessControlManager",
        "type": "address"
      }
    ],
    "name": "NewAccessControlManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferStarted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "mantissa",
            "type": "uint256"
          }
        ],
        "indexed": false,
        "internalType": "struct ExponentialNoError.Exp",
        "name": "marketBorrowIndex",
        "type": "tuple"
      }
    ],
    "name": "RewardTokenBorrowIndexUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newSpeed",
        "type": "uint256"
      }
    ],
    "name": "RewardTokenBorrowSpeedUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "RewardTokenGranted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "RewardTokenSupplyIndexUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract VToken",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "newSpeed",
        "type": "uint256"
      }
    ],
    "name": "RewardTokenSupplySpeedUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint32",
        "name": "newBlock",
        "type": "uint32"
      }
    ],
    "name": "SupplyLastRewardingBlockUpdated",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "INITIAL_INDEX",
    "outputs": [
      {
        "internalType": "uint224",
        "name": "",
        "type": "uint224"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "acceptOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "accessControlManager",
    "outputs": [
      {
        "internalType": "contract IAccessControlManagerV8",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "holder",
        "type": "address"
      },
      {
        "internalType": "contract VToken[]",
        "name": "vTokens",
        "type": "address[]"
      }
    ],
    "name": "claimRewardToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "holder",
        "type": "address"
      }
    ],
    "name": "claimRewardToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "mantissa",
            "type": "uint256"
          }
        ],
        "internalType": "struct ExponentialNoError.Exp",
        "name": "marketBorrowIndex",
        "type": "tuple"
      }
    ],
    "name": "distributeBorrowerRewardToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "supplier",
        "type": "address"
      }
    ],
    "name": "distributeSupplierRewardToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getBlockNumber",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "grantRewardToken",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract Comptroller",
        "name": "comptroller_",
        "type": "address"
      },
      {
        "internalType": "contract IERC20Upgradeable",
        "name": "rewardToken_",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "loopsLimit_",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "accessControlManager_",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "initializeMarket",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "lastContributorBlock",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "maxLoopsLimit",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pendingOwner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "rewardToken",
    "outputs": [
      {
        "internalType": "contract IERC20Upgradeable",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenAccrued",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenBorrowSpeeds",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenBorrowState",
    "outputs": [
      {
        "internalType": "uint224",
        "name": "index",
        "type": "uint224"
      },
      {
        "internalType": "uint32",
        "name": "block",
        "type": "uint32"
      },
      {
        "internalType": "uint32",
        "name": "lastRewardingBlock",
        "type": "uint32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenBorrowerIndex",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenContributorSpeeds",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenSupplierIndex",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenSupplySpeeds",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "rewardTokenSupplyState",
    "outputs": [
      {
        "internalType": "uint224",
        "name": "index",
        "type": "uint224"
      },
      {
        "internalType": "uint32",
        "name": "block",
        "type": "uint32"
      },
      {
        "internalType": "uint32",
        "name": "lastRewardingBlock",
        "type": "uint32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "accessControlManager_",
        "type": "address"
      }
    ],
    "name": "setAccessControlManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "contributor",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "rewardTokenSpeed",
        "type": "uint256"
      }
    ],
    "name": "setContributorRewardTokenSpeed",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken[]",
        "name": "vTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint32[]",
        "name": "supplyLastRewardingBlocks",
        "type": "uint32[]"
      },
      {
        "internalType": "uint32[]",
        "name": "borrowLastRewardingBlocks",
        "type": "uint32[]"
      }
    ],
    "name": "setLastRewardingBlocks",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "limit",
        "type": "uint256"
      }
    ],
    "name": "setMaxLoopsLimit",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract VToken[]",
        "name": "vTokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "supplySpeeds",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256[]",
        "name": "borrowSpeeds",
        "type": "uint256[]"
      }
    ],
    "name": "setRewardTokenSpeeds",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "contributor",
        "type": "address"
      }
    ],
    "name": "updateContributorRewards",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      },
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "mantissa",
            "type": "uint256"
          }
        ],
        "internalType": "struct ExponentialNoError.Exp",
        "name": "marketBorrowIndex",
        "type": "tuple"
      }
    ],
    "name": "updateRewardTokenBorrowIndex",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "vToken",
        "type": "address"
      }
    ],
    "name": "updateRewardTokenSupplyIndex",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
""")
        


from query_state_lib.base.mappers.eth_call_mapper import EthCall
from web3 import Web3, contract

from defi_services.constants.token_constant import Token, ProtocolNFT
from defi_services.utils.logger_utils import get_logger

logger = get_logger('Batch queries')

w3 = Web3()


def add_rpc_call(abi, fn_name, contract_address, block_number=None, fn_paras=None, list_rpc_call=None,
                 list_call_id=None, call_id=None):
    args = []
    if fn_paras is not None:
        if type(fn_paras) is list:
            args = fn_paras
        else:
            if Web3.is_address(fn_paras):
                fn_paras = Web3.to_checksum_address(fn_paras)
            args = [fn_paras]

        if call_id is None:
            call_id = f"{fn_name}_{contract_address}_{fn_paras}_{block_number}".lower()
    else:
        if call_id is None:
            call_id = f"{fn_name}_{contract_address}_{block_number}".lower()

    if call_id in list_call_id:
        return

    c = contract.Contract
    c.w3 = w3
    c.abi = abi
    data_call = c.encodeABI(fn_name=fn_name, args=args)

    if block_number:
        eth_call = EthCall(to=Web3.to_checksum_address(contract_address), block_number=block_number, data=data_call,
                           abi=abi, fn_name=fn_name, id=call_id)
    else:
        eth_call = EthCall(to=Web3.to_checksum_address(contract_address), data=data_call,
                           abi=abi, fn_name=fn_name, id=call_id)

    list_rpc_call.append(eth_call)
    list_call_id.append(call_id)


def decode_data_response(data_responses, list_call_id):
    decoded_datas = {}
    for call_id in list_call_id:
        try:
            response_data = data_responses.get(call_id)
            decoded_data = response_data.decode_result()
        except Exception as e:
            logger.error(f"An exception when decode data from provider: {e}")
            raise

        if len(decoded_data) == 1:
            decoded_datas[call_id] = decoded_data[0]
        else:
            decoded_datas[call_id] = decoded_data
    return decoded_datas


def decode_data_response_ignore_error(data_responses, list_call_id):
    decoded_datas = {}
    for call_id in list_call_id:
        response_data = data_responses.get(call_id)
        try:
            decoded_data = response_data.decode_result()
        except OverflowError:
            result = response_data.result
            if result.startswith('0x'):
                result = result[2:]
            bytes_array = bytearray.fromhex(result)
            bytes32 = bytes_array.hex().rstrip("0")
            if len(bytes32) % 2 != 0:
                bytes32 = bytes32 + '0'
            decoded_data = bytes.fromhex(bytes32).decode('utf8')
        except Exception as e:
            decoded_datum = check_data(call_id, response_data)
            if decoded_datum is not None:
                decoded_datas[call_id] = decoded_datum
                continue
            logger.error(f"An exception when decode data from provider: {e}")
            continue

        if len(decoded_data) == 1:
            decoded_datas[call_id] = decoded_data[0]
        else:
            decoded_datas[call_id] = decoded_data
    return decoded_datas


def check_data(call_id: str, data):
    keys = call_id.split("_")
    fn = keys[0]
    if fn == "underlying" and data.result == "0x":
        return Token.native_token
    if fn == "decimals" and keys[1] in ProtocolNFT.nft:
        return 0
    return None



import logging

from defi_services.abis.dex.pancakeswap.masterchef_v0_abi import PANCAKESWAP_MASTERCHEF_V0_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.dex_constant import Dex
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.dex.dex_info.pancakeswap_info import PANCAKESWAP_V0_BSC_INFO
from defi_services.services.dex.pancakeswap_v2_service import PancakeSwapV2Services

logger = logging.getLogger("PancakeSwap Staking V1 Pool State Service")


class PancakeSwapInfo:
    mapping = {
        Chain.bsc: PANCAKESWAP_V0_BSC_INFO
    }


class PancakeSwapServices(PancakeSwapV2Services):
    def __init__(self, state_service: StateQuerier, chain_id: str = '0x38'):
        super().__init__(state_service=state_service, chain_id=chain_id)

        self.pool_info = PancakeSwapInfo.mapping.get(chain_id)
        self.masterchef_abi = PANCAKESWAP_MASTERCHEF_V0_ABI

    def get_service_info(self):
        info = {
            Dex.pancake: {
                "chain_id": self.chain_id,
                "type": "dex",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_farming_supported_lp_token(self, limit: int = 1):
        web3 = self.state_service.get_w3()
        masterchef_addr = self.pool_info.get('master_chef_address')

        master_chef_contract = web3.eth.contract(abi=self.masterchef_abi,
                                                 address=web3.to_checksum_address(masterchef_addr))
        pool_length = master_chef_contract.functions.poolLength().call()

        rpc_calls = {}
        for pid in range(0, min(pool_length, limit)):
            query_id = f'lpToken_{masterchef_addr}_{pid}_latest'.lower()
            rpc_calls[query_id] = self.get_masterchef_function_info(fn_name="poolInfo", fn_paras=[pid])

        return rpc_calls

    def decode_farming_supported_lp_token(self, decoded_data):
        result = {}
        for query_id, value in decoded_data.items():
            # Format query_id: f'poolInfo_{self.masterchef_addr}_{pid}_latest'

            lp_token = value[0].lower()
            pid = int(query_id.split("_")[-2])
            result[lp_token] = {"farming_pid": pid}

        return result

    # User Information
    def get_user_info_function(
            self, wallet: str, supplied_data, stake: bool = True, block_number: int = "latest"):
        masterchef_addr = self.pool_info.get('master_chef_address')

        rpc_calls = {}

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if stake and (info.get('farming_pid') is not None):
                pid = int(info.get('farming_pid'))

                query_id = f'userInfo_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()
                rpc_calls[query_id] = self.get_masterchef_function_info(
                    fn_name="userInfo", fn_paras=[pid, wallet], block_number=block_number)

        return rpc_calls

    def decode_user_info_function(
            self, wallet: str, supplied_data, decoded_data: dict, stake: bool = True, block_number: int = "latest"):

        result = {}

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if stake and (info.get('farming_pid') is not None):
                pid = int(info.get('farming_pid'))
                query_id = f'userInfo_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()

                user_info = decoded_data.get(query_id)
                stake_amount = user_info[0] / 10 ** info.get('decimals', 18) if user_info else 0

                total_supply = info.get('total_supply')

                result[lp_token] = {
                    'farming_pid': pid,
                    'stake_amount': stake_amount,
                    'tokens': {
                        info['token0']: {
                            'idx': 0,
                            'stake_amount': stake_amount * info.get('token0_amount',
                                                                    0) / total_supply if total_supply else 0
                        },
                        info['token1']: {
                            'idx': 1,
                            'stake_amount': stake_amount * info.get('token1_amount',
                                                                    0) / total_supply if total_supply else 0
                        }
                    }
                }

        return result
